function love.load()
    flow = love.graphics.newPixelEffect [[
        extern number sx;
        extern number sy;

        const float d00 = 8.0 / 8.0;
        const float dr0 = 0.0 / 8.0;
        const float drd = 1.0 / 8.0;
        const float d0d = 2.0 / 8.0;
        const float dld = 3.0 / 8.0;
        const float dl0 = 4.0 / 8.0;
        const float dlu = 5.0 / 8.0;
        const float d0u = 6.0 / 8.0;
        const float dru = 7.0 / 8.0;

        extern float w[9]; 

        vec4 effect(vec4 colour, Image fluid, vec2 grid, vec2 screen)
        {
            vec4 e00 = Texel(fluid, vec2(grid.x   , grid.y   ));
            vec4 er0 = Texel(fluid, vec2(grid.x+sx, grid.y   ));
            vec4 erd = Texel(fluid, vec2(grid.x+sx, grid.y+sy));
            vec4 e0d = Texel(fluid, vec2(grid.x   , grid.y+sy));
            vec4 eld = Texel(fluid, vec2(grid.x-sx, grid.y+sy));
            vec4 el0 = Texel(fluid, vec2(grid.x-sx, grid.y   ));
            vec4 elu = Texel(fluid, vec2(grid.x-sx, grid.y-sy));
            vec4 e0u = Texel(fluid, vec2(grid.x   , grid.y-sy));
            vec4 eru = Texel(fluid, vec2(grid.x+sx, grid.y-sy));

            number weight = w[int(e00.g * 8)];
            number density = e00.r * (1.0 - weight) * 0.5;

            if (abs(el0.g - dr0) < 0.1) density += el0.r * w[1];
            if (abs(elu.g - drd) < 0.1) density += elu.r * w[2];
            if (abs(e0u.g - d0d) < 0.1) density += e0u.r * w[1];
            if (abs(eru.g - dld) < 0.1) density += eru.r * w[2];
            if (abs(er0.g - dl0) < 0.1) density += er0.r * w[1];
            if (abs(erd.g - dlu) < 0.1) density += erd.r * w[2];
            if (abs(e0d.g - d0u) < 0.1) density += e0d.r * w[1];
            if (abs(eld.g - dru) < 0.1) density += eld.r * w[2];

            density += el0.r * w[1] * 0.5 / 8.0;
            density += elu.r * w[2] * 0.5 / 8.0;
            density += e0u.r * w[1] * 0.5 / 8.0;
            density += eru.r * w[2] * 0.5 / 8.0;
            density += er0.r * w[1] * 0.5 / 8.0;
            density += erd.r * w[2] * 0.5 / 8.0;
            density += e0d.r * w[1] * 0.5 / 8.0;
            density += eld.r * w[2] * 0.5 / 8.0;
            //density = min(1.0, density);

            //if (density > 1.0) e00.g = 4/8;

            return vec4(density, e00.g, 1.0, 1.0);
        }
    ]]

    draw = love.graphics.newPixelEffect [[
        vec4 effect(vec4 colour, Image fluid, vec2 grid, vec2 screen)
        {
            vec4 e00 = Texel(fluid, grid);

            return vec4(e00.r, 0.0, 0.0, 1.0);
        }
    ]]

    input = love.graphics.newPixelEffect [[
        vec4 effect(vec4 colour, Image fluid, vec2 grid, vec2 screen)
        {
            vec4 e00 = Texel(fluid, grid);

            return vec4(e00.r, colour.g, e00.b, e00.a);
        }
    ]]

    flow:send("sx", 1 / love.graphics.getWidth())
    flow:send("sy", 1 / love.graphics.getHeight())
    
    local w0 = 0.0
    local w1 = 0.5
    local w2 = math.sqrt(2 * (w1 * w1))

    flow:send("w", w1, w2, w1, w2, w1, w2, w1, w2, w0)

    fluid1 = love.graphics.newCanvas()
    fluid2 = love.graphics.newCanvas()
    --input = love.graphics.newCanvas()

    love.graphics.setColor(0, 3/8* 255, 255, 255)
    fluid2:renderTo(function() love.graphics.rectangle("fill", 0, 0,
                                                       800, 600) end)

    local direction = 1/8
    love.graphics.setColor(128, direction * 255, 255, 255)
    fluid2:renderTo(function() love.graphics.rectangle("fill", 100, 50,
                                                       600, 500) end)

    timer = 0
    tstep = 0.05

    cx, cy = love.mouse.getPosition()

    brush = love.graphics.newQuad(0, 0, 32, 32, love.graphics.getWidth(),
                                                love.graphics.getHeight())
end

function love.update(dt)
    px, py = cx, cy
    cx, cy = love.mouse.getPosition()

    timer = timer + dt

    if timer < tstep then
        return
    else
        timer = timer - tstep
    end

    if love.mouse.isDown("l") then
        local x, y = love.mouse.getPosition()
        brush:setViewport(x, y, 1, 1)

        local direction = math.random(0,7)

        print(direction)

        love.graphics.setColor(255, direction/8 * 255, 255, 255)
        love.graphics.setPixelEffect(input)
        fluid2:renderTo(function() love.graphics.drawq(fluid2, brush, x, y) end)
    end

    love.graphics.setColor(255, 255, 255, 255)

    -- simulate from fluid2 into fluid1
    love.graphics.setPixelEffect(flow)
    fluid1:renderTo(function()
                        for i=1,20 do
                            love.graphics.draw(fluid2)
                        end
                        --love.graphics.draw(input)
                    end)

    -- cut and paste fluid1 into fluid2
    love.graphics.setPixelEffect()
    fluid2:clear()
    fluid2:renderTo(function() love.graphics.draw(fluid1) end)
    fluid1:clear()

    data = fluid2:getImageData()
    total = 0

    --[[
    for y=0,love.graphics.getHeight()-1,4 do
        for x=0,love.graphics.getWidth()-1,4 do
            if data:getPixel(x, y) > 64 then
                total = total + 1
            end
        end
    end
    ]]

    -- steps
    -- add input
    -- diffuse 20 times (do this over time?) -- probably not until later
    -- 
end

function love.draw()
    love.graphics.setColor(255, 255, 255, 255)
    love.graphics.setPixelEffect(draw)
    love.graphics.draw(fluid2)

    love.graphics.setPixelEffect()
    love.graphics.print(love.timer.getFPS(), 20, 20)
    love.graphics.print(total, 20, 60)
end
