require "class"

class "Grid" {
    width = 17, height = 13;
    size = 50;
    factor = 0.01;
}

function Grid:__init__()
    self.time = {}
    self.flow = {}

    self.timer = 0

    for y=0,self.height-1 do
        for x=0,self.width-1 do
            local ci = self:index(x, y)

            self.time[ci] = math.random() / 2
            self.flow[ci] = math.random(0, 7)
        end
    end
end

function Grid:update(dt)
    local changes = {}
    
    for i=1,self.height*self.width do
        changes[i] = 0
    end

    local function wrap(n)
        if n >= 8 then n = n - 8 end
        if n <  0 then n = n + 8 end

        return n
    end

    --local f1, f2 = self.factor, 1 - self.factor
    local r2 = 1/math.sqrt(2)
    local f = {1, r2, 1, r2, 1, r2, 1, r2}

    for y=0,self.height-1 do
        for x=0,self.width-1 do
            local ci = self:index(x, y)

            local nx, ny, valid = self:neighbour(x, y, self.flow[ci])
            local ni = self:index(nx, ny)

            local time = self.time[ci]
            local change = time * 0.1 * f[self.flow[ci]+1]

            changes[ci] = changes[ci] - change
            if valid then changes[ni] = changes[ni] + change end

            if not valid then
                self.flow[ci] = math.random(0, 7)
            end

            --if self.time[ci] > 1.0 then
                local change = time * 0.05
                changes[ci] = changes[ci] - change

                for i=0,7 do
                    local ni = self:index(self:neighbour(x, y, i))

                    changes[ni] = changes[ni] + change * 0.125
                end
            --end

            if self.time[ni] > 2.0 then
                self.flow[ci] = 0
                --[[
                local cn = self.flow[ci]
                local ln, rn = wrap(cn-1), wrap(cn+1)
                local li, ri = self:index(self:neighbour(x, y, ln)),
                               self:index(self:neighbour(x, y, rn))

                local lt, rt = self.time[li], self.time[ri]

                if rt < lt then self.flow[ci] = rn
                else            self.flow[ci] = ln end]]
            end

            if x == 0 then
                self.time[ci] = math.min(0.5, self.time[ci] + 1 * dt)
            end

            if x == self.width-1 then
                self.time[ci] = math.max(0, self.time[ci] - 1 * dt)
            end
        end
    end

    for i, v in pairs(changes) do
        self.time[i] = self.time[i] + v
    end
end

function Grid:draw()
    local div = math.pi * 0.25
    local directions = {}

    for i=0,7 do
        directions[i*2+1] = self.size *  math.cos(div*i) * 0.5
        directions[i*2+2] = self.size * -math.sin(div*i) * 0.5
    end

    for y=0,self.height-1 do
        for x=0,self.width-1 do
            local ci = self:index(x, y)
            local x1, y1 = x*self.size, 600-y*self.size

            local d = self.flow[ci]
            local t = self.time[ci]

            --local nx, ny = s
            --local x2, y2 = nx*self.size, 600-ny*self.size

            local dx, dy = directions[d*2+1], directions[d*2+2]

            love.graphics.setColor(0, 192, 255, 255)
            love.graphics.line(x1, y1, x1+dx, y1+dy)

            love.graphics.setColor(0, 160, 255, 255)
            love.graphics.line(x1, y1, x1+dx/2, y1+dy/2)
        end
    end

    love.graphics.circle("fill", cx*self.size, 600-cy*self.size, self.size*0.5, 16)
end

function Grid:bound(x, y)
    if x < 0 or x >= self.width or y < 0 or y >= self.height then
        return math.max(0, math.min(x, self.width-1)),
               math.max(0, math.min(y, self.height-1)), false
    else
        return x, y, true
    end
end

function Grid:point(x, y)
    return math.floor(x / self.size + 0.5),
           math.floor((600 - y) / self.size + 0.5)
end

function Grid:index(x, y)
    return 1 + y * self.width + x
end

function Grid:neighbour(x, y, i)
    if     i == 0 then
        x, y = x+1, y
    elseif i == 1 then
        x, y = x+1, y+1
    elseif i == 2 then
        x, y = x  , y+1
    elseif i == 3 then
        x, y = x-1, y+1
    elseif i == 4 then
        x, y = x-1, y
    elseif i == 5 then
        x, y = x-1, y-1
    elseif i == 6 then
        x, y = x  , y-1
    elseif i == 7 then
        x, y = x+1, y-1
    end

    return Grid:bound(x, y)
end

function Grid:inv_neighbour(cx, cy, nx, ny)
    local dx, dy = nx - cx, ny - cy

    if     dx ==  1 and dy ==  0 then
        return 0
    elseif dx ==  1 and dy ==  1 then
        return 1
    elseif dx ==  0 and dy ==  1 then
        return 2
    elseif dx == -1 and dy ==  1 then
        return 3
    elseif dx == -1 and dy ==  0 then
        return 4
    elseif dx == -1 and dy == -1 then
        return 5
    elseif dx ==  0 and dy == -1 then
        return 6
    elseif dx ==  1 and dy == -1 then
        return 7
    else
        return self.flow[self:index(cx, cy)]
    end
end

function love.load()
    shade = love.graphics.newPixelEffect([[
        vec3 HSV(float h, float s, float v)
        {
            if (s <= 0 ) { return vec3 (v); }
            h = h * 6;
            float c = v*s;
            float x = (1-abs((mod(h,2)-1)))*c;
            float m = v-c;
            float r = 0.0;
            float g = 0.0;
            float b = 0.0;

            if (h < 1) { r = c; g = x;b = 0.0;}
            else if (h < 2) { r = x; g = c; b = 0.0; }
            else if (h < 3) { r = 0.0; g = c; b = x; }
            else if (h < 4) { r = 0.0; g = x; b = c; }
            else if (h < 5) { r = x; g = 0.0; b = c; }
            else  { r = c; g = 0.0; b = x; }

            return vec3(r+m,g+m,b+m);
        }

        extern number cells[2806];
        extern number width;
        extern number length;
        extern number size;

        vec4 effect(vec4 colour, Image texture, vec2 tex, vec2 fragment)
        {
            int w = int(width);
            int x = int((fragment.x) / size);
            int y = int((fragment.y) / size);

            vec2 p0 = vec2(   x *size,    y *size);
            vec2 p1 = vec2((x+1)*size,    y *size);
            vec2 p2 = vec2((x+1)*size, (y+1)*size);
            vec2 p3 = vec2(   x *size, (y+1)*size);

            int c0 =  y   *w +    x ;
            int c1 =  y   *w + (x+1);
            int c2 = (y+1)*w + (x+1);
            int c3 = (y+1)*w +    x ;

            number v0 = cells[c0];
            number v1 = cells[c1];
            number v2 = cells[c2];
            number v3 = cells[c3];

            if (c0 < 0 || c0 > length-1) v0 = 0.5;
            if (c1 < 0 || c1 > length-1) v1 = 0.5;
            if (c2 < 0 || c2 > length-1) v2 = 0.5;
            if (c3 < 0 || c3 > length-1) v3 = 0.5;

            number dx = 1.0 - (p1.x - fragment.x) / size;
            number dy = 1.0 - (p3.y - fragment.y) / size;

            number vl = v0 + (v3 - v0)*dy;
            number vr = v1 + (v2 - v1)*dy;
            number vm = vl + (vr - vl)*dx;

            number time = vm;

            vec4 hsva = Texel(texture, tex); // * colour;
            vec3 rgb = HSV(hsva.r, hsva.g * time, hsva.b);
            vec4 rgba = vec4(rgb.r, rgb.g, rgb.b, hsva.a);

            return rgba;
        }
    ]])

    grid = Grid()

    level = love.graphics.newCanvas()
    level:renderTo(draw)
    screen = love.graphics.newCanvas()

    shade:send("size", grid.size)
    shade:send("width", grid.width)
    shade:send("length", grid.width * grid.height)
    love.graphics.setPixelEffect(shade)
end

function love.update(dt)
    px, py = cx, cy
    cx, cy = grid:point(love.mouse.getPosition())

    if px and py and cx and cy and love.mouse.isDown("l") then
        grid.flow[grid:index(px, py)] = grid:inv_neighbour(px, py, cx, cy)
    end

    grid:update(dt)
    shade:send("cells", unpack(grid.time))
end

function draw()
    love.graphics.setColor(128, 192, 192, 255)
    love.graphics.rectangle("fill", 0, 0, 800, 600)

    love.graphics.setBlendMode("alpha")
    love.graphics.setColor(255 * math.random(), 255, 128, 255)
    love.graphics.rectangle("fill", 0, 0, 800, 600)

    for i=1,50 do
        local hue = 255 * math.random()
        local sat, val = 255, 128

        local x, y = 800 * math.random(), 600 * math.random()
        local radius = 50 + 30 * math.random()

        local r, g, b = hue, sat, val--HSV(hue, sat, val)

        love.graphics.setColor(r, g, b, 255)
        love.graphics.circle("fill", x, y, radius, 64)
    end
end

function love.draw()
    love.graphics.setPixelEffect()
    screen:renderTo(function () --love.graphics.draw(level, 0, 0)
                                love.graphics.setColor(128, 255, 128, 255)
                                love.graphics.rectangle("fill", 0, 0, 800, 600)
                                grid:draw() end)

    love.graphics.setPixelEffect(shade)
    love.graphics.draw(screen, 0, 0)

    --love.graphics.setPixelEffect()
    --grid:draw()

    love.graphics.setColor(255, 255, 255, 255)
    love.graphics.print("Current FPS: "..tostring(love.timer.getFPS()),
                        10, 10)
end
